import numpy as np
import pandas as pd


def ht(arr, n: int = 2):
    """Get first and last (top and bottom) entries

    :param arr: DataFrame, Series, array, list, or tuple
    :param n: number of top and bottom entries
    :return: concatenated top and bottom entries
    """

    if len(arr) <= n * 2:
        return arr
    elif isinstance(arr, pd.DataFrame) or isinstance(arr, pd.Series):
        return pd.concat([arr.head(n), arr.tail(n)])
    elif isinstance(arr, np.ndarray):
        return np.concatenate([arr[:n], arr[-n:]])
    elif isinstance(arr, list) or isinstance(arr, tuple):
        return arr[:n] + arr[-n:]
    else:
        raise ValueError(f'{arr} type is not recognized')



from pathlib import Path
import dask
from dask.delayed import Delayed
from typing import Optional, Union

from dutil.pipeline import cached, CachedResult


def cached_delayed(
    name: Optional[str] = None,
    name_prefix: Optional[str] = None,
    parameters: Optional[dict] = None,
    ignore_args_kwargs: Optional[bool] = None,
    folder: Union[str, Path] = 'cache',
    ftype: str = 'pickle',
    override: bool = False,
    logger=None,
):
    """cached + dask.delayed
    
    Returns Delayed object with smart output caching
    """

    def decorator(foo):
        cached_foo = cached(
            name=name,
            name_prefix=name_prefix,
            parameters=parameters,
            ignore_args_kwargs=ignore_args_kwargs,
            folder=folder,
            ftype=ftype,
            override=override,
            logger=logger,
        )(foo)
        new_foo = dask.delayed()(cached_foo)
        return new_foo
    return decorator


class DelayedParameters():
    """Delayed parameters
    
    Important! Method `set` does not work with dask distributed.
    """

    def __init__(self):
        self._params = {}
        
    def new(self, name: str, value=None) -> Delayed:
        """Return a delayed object for the new parameter"""
        if name in self._params:
            raise KeyError(f'Parameter {name} already exists')
        self._params[name] = value
        return dask.delayed(name=name)(lambda: self._params[name])()
        
    def update(self, name: str, value) -> None:
        """Update parameter value"""
        if name not in self._params:
            raise KeyError(f'Parameter {name} does not exist')
        self._params[name] = value

    def update_many(self, d: dict) -> None:
        """Update multiple parameters values"""
        for k, v in d.items():
            self.update(k, v)


def dask_compute(tasks, scheduler='threads') -> tuple:
    """Compute values of Delayed objects"""
    results = dask.compute(*tasks, scheduler=scheduler)
    datas = tuple(r.load() if isinstance(r, CachedResult) else r for r in results)
    return datas



import pyarrow
import functools
# from contextlib import contextmanager
# from contextvars import ContextVar
import numpy as np
import pandas as pd
import dill
from pathlib import Path
import shutil
import xxhash
from dask.delayed import Delayed
from typing import Optional, Union
from loguru import logger as _logger
import json
import warnings

_ = pyarrow.__version__  # explicitly show pyarrow dependency

xxhasher = xxhash.xxh64(seed=42)


def _hash_obj(obj):
    if isinstance(obj, CachedResult):
        h = obj.get_hash()
    elif isinstance(obj, np.ndarray):
        xxhasher.update(obj.data)
        h = str(xxhasher.intdigest())
        xxhasher.reset()
    elif isinstance(obj, pd.Series):
        xxhasher.update(obj.values.data)
        h = str(xxhasher.intdigest())
        xxhasher.reset()
    elif isinstance(obj, pd.DataFrame):
        for c in obj:
            try:
                xxhasher.update(obj[c].values.data)
            except ValueError:
                xxhasher.update(obj[c].astype(str).values.data)
        h = str(xxhasher.intdigest())
        xxhasher.reset()
    else:
        h = str(obj)
    return h


def _get_cache_path(name, name_prefix, parameters, ignore_args_kwargs,
                    folder, ftype, foo, args, kwargs) -> Path:
    path = Path(folder)
    if ignore_args_kwargs is None:
        ignore_args_kwargs = (parameters is not None)
    if name is None:
        _n = [foo.__name__, ]
        if parameters is not None:
            _n.extend([str(k) + _hash_obj(v) for k, v in parameters.items()])
        if not ignore_args_kwargs:
            _n.extend([_hash_obj(a) for a in args])
            _n.extend([str(k) + _hash_obj(v) for k, v in kwargs.items()])
        name = '_'.join(_n) + '.' + ftype
    if name_prefix is not None:
        name = name_prefix + '__' + name
    path = path / name
    return path


def _cached_load(ftype, path):
    if ftype == 'parquet':
        data = pd.read_parquet(path)
    elif ftype == 'pickle':
        data = dill.load(open(path, 'rb'))
    else:
        raise ValueError('ftype {} is not recognized'.format(ftype))
    return data


def _cached_save(data, ftype, path) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    if ftype == 'parquet':
        data.to_parquet(path, index=False, allow_truncated_timestamps=True)
    elif ftype == 'pickle':
        dill.dump(data, open(path, 'wb'))
    else:
        raise ValueError('ftype {} is not recognized'.format(ftype))


def cached0(
    name: Optional[str] = None,
    name_prefix: Optional[str] = None,
    parameters: Optional[dict] = None,
    ignore_args_kwargs: Optional[bool] = None,
    folder: Union[str, Path] = 'cache',
    ftype: str = 'pickle',
    override: bool = False,
    logger=None,
):
    """Cache function output on the disk

    Features:
    - Pickle and parquet serialization
    - Special treatment for Delayed objects

    :param name: name of the cache file
        if none, name is constructed from the function name and args
    :param parameters: include these parameters in the name
        only meaningful when `name=None`
    :param ignore_args_kwargs: if true, do not add args and kwargs to the name
        only meaningful when `name=None`
    :param folder: name of the cache folder
    :param ftype: type of the cache file
        'pickle' | 'parquet'
    :param override: if true, override the existing cache file
    :param logger: if none, use a new logger
    :return: new function
        output is loaded from cache file if it exists, generated otherwise
    """

    warnings.warn('This function will be depreciated', category=DeprecationWarning)
    logger = logger if logger is not None else _logger

    def decorator(foo):
        @functools.wraps(foo)
        def new_foo(*args, **kwargs):
            path = _get_cache_path(name, name_prefix, parameters, ignore_args_kwargs,
                                   folder, ftype, foo, args, kwargs)
            if not override and path.exists():
                data = _cached_load(ftype, path)
                logger.debug('data has been loaded from {}'.format(path))
            else:
                data = foo(*args, **kwargs)
                dask_args_detected = any(isinstance(a, Delayed) for a in args)
                dask_kwargs_detected = any(isinstance(v, Delayed) for k, v in kwargs.items())
                if not dask_args_detected and not dask_kwargs_detected:
                    _cached_save(data, ftype, path)
                logger.debug('data has been generated and saved in {}'.format(path))
            return data
        return new_foo
    return decorator


def _get_meta_path(cache_path: Path) -> Path:
    return cache_path.parent / (cache_path.name + '.meta')


class CacheMeta:
    """Cache meta data (incl. hash)"""

    def __init__(
        self,
        name: str,
        meta_path: Union[str, Path],
        cache_path: Union[str, Path],
        ftype: str,
        hash_value: Optional[str]
    ):
        self.name = name
        self.meta_path = Path(meta_path)
        self.cache_path = Path(cache_path)
        self.ftype = ftype
        self.hash_value = hash_value

    @classmethod
    def from_file(cls, meta_path: Path):
        with open(meta_path, 'rt') as f:
            fields = json.load(f)
        return cls(**fields)

    def dump_to_file(self):
        with open(self.meta_path, 'wt') as f:
            fields = {k: str(v) for k, v in self.__dict__.items()}
            json.dump(fields, f)


class CachedResult:
    """Lazy loader for cache data"""

    def __init__(self, name, name_prefix, parameters, ignore_args_kwargs,
                 folder, ftype, foo, args, kwargs, logger):
        cache_path = _get_cache_path(name, name_prefix, parameters, ignore_args_kwargs,
                                     folder, ftype, foo, args, kwargs)
        meta_path = _get_meta_path(cache_path)
        name = cache_path.name
        if meta_path.exists():
            self.meta = CacheMeta.from_file(meta_path)
        else:
            self.meta = CacheMeta(name, meta_path, cache_path, ftype, hash_value=None)
        self._cache_value = None
        self.logger = logger

    def load(self):
        """Load data from cache"""

        if self._cache_value is None:
            self._cache_value = _cached_load(self.meta.ftype, self.meta.cache_path)
            self.logger.debug('Task {}: data has been loaded from cache'.format(self.meta.name))
        return self._cache_value

    def dump(self, data):
        """Dump data to cache"""

        self._cache_value = data
        _cached_save(data, self.meta.ftype, self.meta.cache_path)
        self.logger.debug('Task {}: data has been saved to cache'.format(self.meta.name))
        self.meta.dump_to_file()

    def get_hash(self):
        """Get hash of cached data

        Used to construct a cache file name
        """

        # Hash may not be required, so it's not automatically computed from data
        if self.meta.hash_value is None:
            cache_obj = self.load()
            self.meta.hash_value = _hash_obj(cache_obj)
            self.meta.dump_to_file()
            self.logger.debug('Task {}: hash has been computed from data'.format(self.meta.name))
        return self.meta.hash_value

    def exists(self):
        return self.meta.cache_path.exists() and self.meta.meta_path.exists()


def cached(
    name: Optional[str] = None,
    name_prefix: Optional[str] = None,
    parameters: Optional[dict] = None,
    ignore_args_kwargs: Optional[bool] = None,
    folder: Union[str, Path] = 'cache',
    ftype: str = 'pickle',
    override: bool = False,
    logger=None,
):
    """Cache function output on the disk (advanced version)

    Features:
    - Pickle and parquet serialization
    - Special treatment for Delayed objects
    - Lazy cache loading
    - Hashing of complex arguments

    :param name: name of the cache file
        if none, name is constructed from the function name and args
    :param parameters: include these parameters in the name
        only meaningful when `name=None`
    :param ignore_args_kwargs: if true, do not add args and kwargs to the name
        only meaningful when `name=None`
    :param folder: name of the cache folder
    :param ftype: type of the cache file
        'pickle' | 'parquet'
    :param override: if true, override the existing cache file
    :param logger: if none, use a new logger
    :return: new function
        output is lazily loaded from cache file if it exists, generated otherwise
        .load() to get data
    """

    logger = logger if logger is not None else _logger

    def decorator(foo):
        @functools.wraps(foo)
        def new_foo(*args, **kwargs):
            result = CachedResult(name, name_prefix, parameters,
                                  ignore_args_kwargs,
                                  folder, ftype, foo, args, kwargs,
                                  logger=logger)
            if not override and result.exists():
                # if the result (= cache OR cache + hash) exists, do nothing - just pass it on
                # the cache will be loaded only if required later
                logger.info('Task {}: skip (cache exists)'.format(result.meta.name))
            else:
                # if the result does not exist, generate data and save cache
                dask_args_detected = any(isinstance(a, Delayed) for a in args)
                dask_kwargs_detected = any(isinstance(v, Delayed) for k, v in kwargs.items())
                if not dask_args_detected and not dask_kwargs_detected:
                    # eager load cache for all arguments
                    args = [a.load() if isinstance(a, CachedResult) else a for a in args]
                    kwargs = {k: v.load() if isinstance(v, CachedResult) else v for k, v in kwargs.items()}
                    data = foo(*args, **kwargs)
                    result.dump(data)
                    logger.info('Task {}: data has been computed and saved to cache'.format(result.meta.name))
                else:
                    # if any of the arguments is a Delayed object, return anything
                    result = foo(*args, **kwargs)
            return result
        return new_foo
    return decorator


def clear_cache(
    folder: Union[str, Path] = 'cache',
    ignore_errors: bool = True,
):
    """Clear the cache folder

     :param folder: name of the cache folder
    """
    folder = Path(folder).absolute()
    shutil.rmtree(folder, ignore_errors=ignore_errors)



"""
Data persistance and pipelining tools
"""

from dutil.pipeline._cached import cached0, cached, CachedResult, clear_cache  # noqa: F401
from dutil.pipeline._dask import cached_delayed, DelayedParameters, dask_compute   # noqa: F401



import pytest
import numpy as np
import pandas as pd

from dutil.stats import mean_upper, mean_lower


@pytest.mark.parametrize(
    'data, expected',
    [
        (np.array([0, 1, 5, -1]), 3.),
        (pd.Series([0, 1, 5, -1]), 3.),
    ]
)
def test_mean_upper_assert_equal(data, expected):
    actual = mean_upper(data)
    assert actual == expected


@pytest.mark.parametrize(
    'data, expected',
    [
        (np.array([0, 1, 5, -1]), -.5),
        (pd.Series([0, 1, 5, -1]), -.5),
    ]
)
def test_mean_lower_assert_equal(data, expected):
    actual = mean_lower(data)
    assert actual == expected



import numpy as np
import pandas as pd
from typing import Union


def mean_upper(arr: Union[pd.Series, np.ndarray]) -> float:
    arr = np.asarray(arr)
    return np.mean(np.sort(arr)[len(arr) // 2:])


def mean_lower(arr: Union[pd.Series, np.ndarray]) -> float:
    arr = np.asarray(arr)
    return np.mean(np.sort(arr)[:len(arr) // 2])


def q_01(arr: Union[pd.Series, np.ndarray]) -> float:
    arr = np.asarray(arr)
    return np.quantile(arr, .01)


def q_10(arr: Union[pd.Series, np.ndarray]) -> float:
    arr = np.asarray(arr)
    return np.quantile(arr, .1)


def q_25(arr: Union[pd.Series, np.ndarray]) -> float:
    arr = np.asarray(arr)
    return np.quantile(arr, .25)


def q_50(arr: Union[pd.Series, np.ndarray]) -> float:
    arr = np.asarray(arr)
    return np.quantile(arr, .5)


def q_75(arr: Union[pd.Series, np.ndarray]) -> float:
    arr = np.asarray(arr)
    return np.quantile(arr, .75)


def q_90(arr: Union[pd.Series, np.ndarray]) -> float:
    arr = np.asarray(arr)
    return np.quantile(arr, .90)


def q_99(arr: Union[pd.Series, np.ndarray]) -> float:
    arr = np.asarray(arr)
    return np.quantile(arr, .99)



import pytest
import numpy as np
import pandas as pd
import datetime as dt
import time
from dask import delayed

from dutil.pipeline import cached0, cached, clear_cache, cached_delayed, DelayedParameters, dask_compute


cache_dir = 'cache/temp/'
eps = 0.00001


@pytest.mark.parametrize(
    'data, ftype',
    [
        ((0, 1, 3, 5, -1), 'pickle'),
        ((0, 1., 3232.22, 5., -1., None), 'pickle'),
        ([0, 1, 3, 5, -1], 'pickle'),
        ([0, 1., 3232.22, 5., -1., None], 'pickle'),
        (pd.Series([0, 1, 3, 5, -1]), 'pickle'),
        (pd.Series([0, 1., 3232.22, 5., -1., np.nan]), 'pickle'),
        (pd.DataFrame({
            'a': [0, 1, 3, 5, -1],
            'b': [2, 1, 0, 0, 14],
        }), 'pickle'),
        (pd.DataFrame({
            'a': [0, 1., 3232.22, -1., np.nan],
            'b': ['a', 'b', 'c', 'ee', '14'],
            'c': [dt.datetime(2018, 1, 1),
                  dt.datetime(2019, 1, 1),
                  dt.datetime(2020, 1, 1),
                  dt.datetime(2021, 1, 1),
                  dt.datetime(2022, 1, 1)],
        }), 'pickle'),
        (pd.DataFrame({
            'a': [0, 1, 3, 5, -1],
            'b': [2, 1, 0, 0, 14],
        }), 'parquet'),
        (pd.DataFrame({
            'a': [0, 1., 3232.22, -1., np.nan],
            'b': ['a', 'b', 'c', 'ee', '14'],
            'c': [dt.datetime(2018, 1, 1),
                  dt.datetime(2019, 1, 1),
                  dt.datetime(2020, 1, 1),
                  dt.datetime(2021, 1, 1),
                  dt.datetime(2022, 1, 1)],
            'd': [pd.Timestamp('2018-01-01'),
                  pd.Timestamp('2018-01-01'),
                  pd.Timestamp('2018-01-01'),
                  pd.Timestamp('2018-01-01'),
                  pd.Timestamp('2018-01-01')],
        }), 'parquet'),
        (pd.DataFrame({
            'a': [0, 1., 3232.22, -1., np.nan],
            'b': [dt.timedelta(hours=1),
                  dt.timedelta(hours=1),
                  dt.timedelta(hours=1),
                  dt.timedelta(hours=1),
                  dt.timedelta(hours=1)],
        }), 'pickle'),
    ]
)
def test_cached_assert_equal(data, ftype):
    @cached0(folder=cache_dir, ftype=ftype, override=False)
    def load_data():
        return data

    clear_cache(cache_dir)
    _ = load_data()
    loaded = load_data()

    if isinstance(data, pd.Series):
        pd.testing.assert_series_equal(loaded, data)
    elif isinstance(data, pd.DataFrame):
        pd.testing.assert_frame_equal(loaded, data)
    elif isinstance(data, np.ndarray):
        np.testing.assert_equal(loaded, data)
    else:
        assert loaded == data


@pytest.mark.parametrize(
    'data, ftype, eps, ts',
    [
        (pd.DataFrame({
            'a': [0, 1, 3, 5, -1],
            'b': [2, 1, 0, 0, 14],
        }), 'parquet', 0.1, pd.Timestamp('2018-01-01'),),
        (pd.DataFrame({
            'a': [0, 1., 3232.22, -1., np.nan],
            'b': ['a', 'b', 'c', 'ee', '14'],
            'c': [dt.datetime(2018, 1, 1),
                  dt.datetime(2019, 1, 1),
                  dt.datetime(2020, 1, 1),
                  dt.datetime(2021, 1, 1),
                  dt.datetime(2022, 1, 1)],
            'd': [pd.Timestamp('2018-01-01'),
                  pd.Timestamp('2018-01-01'),
                  pd.Timestamp('2018-01-01'),
                  pd.Timestamp('2018-01-01'),
                  pd.Timestamp('2018-01-01')],
        }), 'parquet', 0.1, pd.Timestamp('2018-01-01'),),
        (pd.DataFrame({
            'a': [0, 1., 3232.22, -1., np.nan],
            'b': [dt.timedelta(hours=1),
                  dt.timedelta(hours=1),
                  dt.timedelta(hours=1),
                  dt.timedelta(hours=1),
                  dt.timedelta(hours=1)],
        }), 'pickle', 0.1, pd.Timestamp('2018-01-01'),),
    ]
)
def test_cached_with_args_kwargs_assert_equal(data, ftype, eps, ts):
    @cached0(folder=cache_dir, ftype=ftype, override=False)
    def load_data(eps, ts):
        assert eps > 0
        assert ts > pd.Timestamp('2000-01-01')
        return data

    clear_cache(cache_dir)
    _ = load_data(eps, ts=ts)
    loaded = load_data(eps, ts=ts)

    if isinstance(data, pd.Series):
        pd.testing.assert_series_equal(loaded, data)
    elif isinstance(data, pd.DataFrame):
        pd.testing.assert_frame_equal(loaded, data)
    elif isinstance(data, np.ndarray):
        np.testing.assert_equal(loaded, data)
    else:
        assert loaded == data


@pytest.mark.parametrize(
    'data, output, ftype',
    [
        (pd.DataFrame({
            'a': [0, 1, 3, 5, -1],
            'b': [2, 1, 0, 0, 14],
        }), pd.DataFrame({
            'a': [0, 1, 3, 5, -1],
            'b': [2, 1, 0, 0, 14],
        }), 'parquet'),
        (pd.DataFrame({
            'a': [.5, np.nan, np.nan],
            'b': ['a', 'b', '14'],
            'c': [dt.datetime(2018, 1, 1),
                  dt.datetime(2019, 1, 1),
                  dt.datetime(2022, 1, 1)],
            'd': [pd.Timestamp('2018-01-01'),
                  pd.Timestamp('2018-01-01'),
                  pd.Timestamp('2018-01-01')],
        }), pd.DataFrame({
            'a': [.5],
            'b': ['a'],
            'c': [dt.datetime(2018, 1, 1)],
            'd': [pd.Timestamp('2018-01-01')],
        }), 'parquet'),
        (pd.DataFrame({
            'a': [0, 1., np.nan],
            'b': [dt.timedelta(hours=1),
                  dt.timedelta(hours=1),
                  dt.timedelta(hours=1)],
        }), pd.DataFrame({
            'a': [0, 1.],
            'b': [dt.timedelta(hours=1),
                  dt.timedelta(hours=1)],
        }), 'pickle'),
    ]
)
def test_cached_with_chained_df_assert_equal(data, output, ftype):
    @cached0(folder=cache_dir, ftype=ftype, override=False)
    def load_data():
        return data
    
    @cached0(folder=cache_dir, ftype=ftype, override=False)
    def process_data(df):
        return df.dropna()

    clear_cache(cache_dir)
    df = load_data()
    _ = process_data(df)

    df = load_data()
    processed = process_data(df)

    if isinstance(data, pd.Series):
        pd.testing.assert_series_equal(processed, output)
    elif isinstance(data, pd.DataFrame):
        pd.testing.assert_frame_equal(processed, output)
    elif isinstance(data, np.ndarray):
        np.testing.assert_equal(processed, output)
    else:
        assert processed == output


@pytest.mark.parametrize(
    'data, ftype, eps, ts',
    [
        (pd.DataFrame({
            'a': [0, 1, 3, 5, -1],
            'b': [2, 1, 0, 0, 14],
        }), 'parquet', 0.1, pd.Timestamp('2018-01-01'),),
    ]
)
def test_dask_cached_with_args_kwargs_assert_equal(data, ftype, eps, ts):
    @delayed()
    @cached(folder=cache_dir, ftype=ftype, override=False)
    def load_data(eps, ts):
        assert eps > 0
        assert ts > pd.Timestamp('2000-01-01')
        return data

    clear_cache(cache_dir)
    r = load_data(eps, ts=ts)
    _ = r.compute()
    loaded = r.compute().load()

    if isinstance(data, pd.Series):
        pd.testing.assert_series_equal(loaded, data)
    elif isinstance(data, pd.DataFrame):
        pd.testing.assert_frame_equal(loaded, data)
    elif isinstance(data, np.ndarray):
        np.testing.assert_equal(loaded, data)
    else:
        assert loaded == data


def test_cached_load_time():
    @cached0(folder=cache_dir, override=False)
    def load_data():
        time.sleep(1)
        return 1

    clear_cache(cache_dir)

    start = dt.datetime.utcnow()
    _ = load_data()
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay > 0.95

    start = dt.datetime.utcnow()
    _ = load_data()
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay < 0.95


def test_dask_cached_load_time():
    @delayed()
    @cached(folder=cache_dir, override=False)
    def load_data():
        time.sleep(1)
        return 1

    clear_cache(cache_dir)

    start = dt.datetime.utcnow()
    r = load_data()
    r.compute()
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay > 0.95

    start = dt.datetime.utcnow()
    r = load_data()
    r.compute()
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay < 0.95


def test_dask_pipeline():
    clear_cache(cache_dir)

    @cached_delayed(folder=cache_dir)
    def load_data_1():
        time.sleep(1)
        return 5

    @cached_delayed(folder=cache_dir)
    def load_data_2():
        time.sleep(1)
        return 3

    @cached_delayed(folder=cache_dir)
    def add(x, y):
        return x + y

    d1 = load_data_1()
    d2 = load_data_2()
    r = add(d1, d2)

    start = dt.datetime.utcnow()
    (output,) = dask_compute((r,))
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert 0.95 < delay < 1.95
    assert output == 8

    start = dt.datetime.utcnow()
    (output,) = dask_compute((r,))
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay < 0.95
    assert output == 8


def test_dask_pipeline_with_parameters():
    clear_cache(cache_dir)

    @cached_delayed(folder=cache_dir)
    def load_data_1(ts: dt.datetime):
        assert ts > dt.datetime(2019, 1, 1)
        time.sleep(1)
        return 5

    @cached_delayed(folder=cache_dir)
    def load_data_2(eps: float):
        time.sleep(1)
        return 3 + eps

    @cached_delayed(folder=cache_dir)
    def add(x, y):
        return x + y

    params = DelayedParameters()
    ts = params.new('ts', value=dt.datetime(2020, 1, 1))
    fix = params.new('fix', value=0.5)
    d1 = load_data_1(ts)
    d2 = load_data_2(fix)
    r = add(d1, d2)

    start = dt.datetime.utcnow()
    (output,) = dask_compute((r,))
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert 0.95 < delay < 1.95
    assert abs(output - 8.5) < eps

    start = dt.datetime.utcnow()
    (output,) = dask_compute((r,))
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay < 0.95
    assert abs(output - 8.5) < eps

    params.update_many({'ts': dt.datetime(2020, 2, 1), 'fix': 1.5})
    start = dt.datetime.utcnow()
    (output,) = dask_compute((r,))
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert 0.95 < delay < 1.95
    assert abs(output - 9.5) < eps

    start = dt.datetime.utcnow()
    (output,) = dask_compute((r,))
    delay = (dt.datetime.utcnow() - start).total_seconds()
    assert delay < 0.95
    assert abs(output - 9.5) < eps



import pathlib
import os
from setuptools import setup, find_packages

PATH = pathlib.Path(__file__).parent
VERSION = os.getenv('VERSION', 'dev')
LONG_DESCRIPTION = (PATH / "README.md").read_text()

setup(
    name='dutil',
    version=VERSION,
    description='A few useful tools for data wrangling',
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author='Yaroslav Kopotilov',
    license='Apache License, Version 2.0',
    author_email='UNKNOWN',
    url='https://github.com/mysterious-ben/dutil',
    install_requires=[
        'numpy',
        'pandas',
        'pyarrow',
        'dill',
        'loguru',
        'xxhash',
        'dask[delayed]',
        'ipython',
    ],
    entry_points={'console_scripts': []},
    packages=find_packages(),
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    python_requires='>=3.7',
    test_suite='tests',
    setup_requires=[
        'pytest-runner',
        'setuptools',
    ],
    tests_require=[
        'pytest',
    ]
)


# --- pip upload ---
# python setup.py sdist bdist_wheel
# twine check dist/*
# twine upload dist/*
